/**
 * Author:  Kevin Chan
 * Company: University of Washington
 *
 * File: 	IrCom.c
 *
 * Purpose: The palm os has several ways of interacting with the IR capabilities
 *			that are inherent in its design. At the lowest level, the palm os
 *			provides direct access to the IrDA implementation. If you want visit
 *			www.irda.org for the actual specification.
 *
 *			While we want access to IR, we do not want to be wrapped up in the
 *			plumbing that is required to get it to work. So this module is provided
 *			to easy IR development.
 *
 *			NOTE: this is a really thin wrapper. You can use some or all of the
 *			wrapper functions, depending on your need.
 */
#include <irlib.h>
#include "IrCom.h"
#include "AppStdIO.h"

UInt refNum = 0;
IrConnect con;

/**
 * DiscTimeout is a random number chosen to allow a tight loop to
 * timeout when querying for values. (XXX Huh, dumb comment will fix later XXX)
 */
const UInt16 DiscTimeout = 300;

// const char lmp_lsap_attr_name[] = "\022IrDA:IrLMP:LsapSel";
// const char tp_lsap_attr_name[]  = "\023IrDA:TinyTP:LsapSel";

/**
 * The following are variables that must be modified. When you change the name
 * you MUST change the corresponding field in irLMPQuery or irTTPQuery.
 *
 * when modifying irLMPQuery or irTTPQuery, MAKE SURE YOU CHANGE the corresponding
 * size as specified by the 2 hex digits preceeding the name.
 */
static const Byte name[] = "IrPing";
static Byte irLMPQuery[] = "\x06IrPing\x12IrDA:IrLMP:LsapSel";
static const Byte irLMPQueryLen = 26;
static Byte irTTPQuery[] = "\x06IrPing\x13IrDA:TinyTP:LsapSel";
static const Byte irTTPQueryLen = 27;

/**
 * Local static variable used to setup the attribute name and to hold
 * the connection type.
 *
 * ias_object and attrib are available for you to look at if you want to
 * directly manipulate them. Highly advise you do not if you use the APIs
 * exposed herein.
 */
static Boolean reqTTPConnection = false;	// Connection type.
static const char attr_ttp_name[] = "IrDA:TinyTP:LsapSel";
static const char attr_lmp_name[] = "IrDA:IrLMP:LsapSel";
static unsigned char  attr_value[]={ IAS_ATTRIB_INTEGER, 0,0,0,42};

IrIasObject ias_object;
IrIasAttribute attrib;

/**
 * Structure used to determine the device info.
 */
Byte XID_DeviceInfo[] = {IR_HINT_PDA, IR_CHAR_ASCII, 'I', 'R', ' ', 'C', 'H', 'A', 'T'};

/**
 * the target device address we want to send packets to. Requires the local boolean
 * value to indicate when the callback function finished correctly.
 */
static Boolean devAddrValid = false;
IrDeviceAddr device_addr;

/**
 * Local variables that indicate the current internal state.
 */
static Boolean irLapSuccess = false;
static Boolean foundObject = false;
static Boolean connected = false;

/**
 * Local query object.
 */
static IrIasQuery iasQuery;
static Boolean gotRLsap = false;
static Byte qRsltBuff[sizeof(UInt32)];
static UInt16 qRsltBuffSize = sizeof(UInt32);

/**
 * Translates IRStaatus values to their textual equivalents.
 */
char* ir_status_messages[] = 
{
	"Successful and complete",
	"Operation failed",
	"Successfully started but pending",
	"Link disconnected",
	"No IrLAP Connection exists",
	"IR Media is busy",
	"IR Media is not busy",
	"IrLAP not making progress",
	"No progress condition cleared",
	NULL
};

char* ir_event_msg [] =
{
 "LM_CON_IND",
 "LM_DISCON_IND",
 "DATA_IND",
 "PACKET_HANDLED",
 "LAP_CON_IND",
 "LAP_DISCON_IND",
 "DISCOVERY_CNF",
 "LAP_CON_CNF",
 "LM_CON_CNF",
 "STATUS_IND",
 "TEST_IND",
 "TEST_CNF",
};


/***********************************************************************
 *
 * FUNCTION:     ir_cb_func
 *
 * DESCRIPTION:  This is an internal callback function provided so that
 * I can get first dibs on require callback plumbing. Anything that I do
 * not handle will be forwarded on to you (i.e. LEVENT_DATA_IND);
 *
 * PARAMETERS:   
 *
 * RETURNED:     0 -- no error
 *				 -1 -- error;
 *
 * REVISION HISTORY:
 *
 ***********************************************************************/
static void ir_cb_func(IrConnect* con, IrCallBackParms* params)
{
	debug_msg_1("event -- %s", ir_event_msg[params->event]);
	
	switch (params->event)
	{
		// Discovery completion operation. This one is fixed.
		// We need the return address. so set the field in IrCom.h
		// appropriately. The only valid params value is in the
		// deviceList member. There always exist one element in this
		// list which is the device object.
		case LEVENT_DISCOVERY_CNF:
		
			// Get the device object's address.
			debug_msg("Received Discovery_CNF");
			device_addr = params->deviceList->dev[0].hDevice;
			devAddrValid = true;
			break;
			
		// The request IrLAP connection was successful.
		case LEVENT_LAP_CON_CNF:
			debug_msg("received LAP_CON_CNF");
			irLapSuccess = true;
			break;
			
		// The requested IrLMP / TinyTP
		// connection was made successfully.
		case LEVENT_LM_CON_CNF:
			debug_msg("received LM_CON_CNF");
			connected = true;
			break;
			
		// This is packet handled events. Is signaled
		// when you successfully processed a packet.
		case LEVENT_PACKET_HANDLED:
			debug_msg("received PACKET_HANDLED");
			break;
			
		// Trap the LEVENT_LM_DISCON_IND to reset internal state
		// but then pass it through to the enduser in case they
		// want to handle it.
		case LEVENT_LM_DISCON_IND:
			connected = false;
			
		default:
			IR_Callback(con, params);
			break;
	}
}

/***********************************************************************
 *
 * FUNCTION:     ir_cb_func
 *
 * DESCRIPTION:  This is an internal callback function provided so that
 * I can get first dibs on require callback plumbing. Anything that I do
 * not handle will be forwarded on to you (i.e. LEVENT_DATA_IND);
 *
 * PARAMETERS:   
 *
 * RETURNED:     0 -- no error
 *				 -1 -- error;
 *
 * REVISION HISTORY:
 *
 ***********************************************************************/
 static void ias_cb_func(IrStatus status)
 {
 	if (status != IR_STATUS_SUCCESS)
 		return;
 		
 	if (iasQuery.retCode != IAS_RET_SUCCESS)
 		return;
 		
 	switch (IrIAS_GetType(&iasQuery))
 	{
 		case IAS_ATTRIB_MISSING:
 			debug_msg("Attribute is missing");
 			return;
 			
 		case IAS_ATTRIB_INTEGER:
 			con.rLsap = IrIAS_GetIntLsap(&iasQuery);
 			debug_msg("Retreved the rLsap");
 			gotRLsap = true;
 	}
 }


/***********************************************************************
 *
 * FUNCTION:     IR_OpenCom
 *
 * DESCRIPTION:  This is a wrapper that handles the logic needed to open
 * a connection to the IR port. It will also register an entry into the
 * IAS to allow you to query for the current service if you want.
 *
 * PARAMETERS:   Boolean connectionless -- Determines whether you use
 *				 LMP (no packet guarantee) or TTP (delivery guarantee)
 *
 * RETURNED:     IR_STATUS_SUCCESS | IR_STATUS_FAILED;
 *
 * REVISION HISTORY:
 *
 ***********************************************************************/
IrStatus IR_Bind( Boolean connectionless )
{
	if (refNum)
		return;
		
	if (SysLibFind( irLibName,&refNum) != 0)
	{
		// Error occurred
		// Handle it
		return IR_STATUS_FAILED;
	}
	
	if (IrOpen(refNum, 0) != 0)
	{
		debug_msg("IrOpenComm::IrOpen failed.");
		return IR_STATUS_FAILED;
	}
	
	
	// Now bind to a local LSAP port
	if (IrBind(refNum, &con, ir_cb_func) != IR_STATUS_SUCCESS)
	{
		debug_msg("IrOpenComm::IrBind failed.");
		return IR_STATUS_FAILED;
	}
	
	if (IrSetDeviceInfo(refNum, XID_DeviceInfo, sizeof(XID_DeviceInfo) != IR_STATUS_SUCCESS))
	{
		debug_msg("IrOpenComm::IrSetDeviceInfo failed.");
		IR_CloseCom();
		return IR_STATUS_FAILED;
	}
	
	// Now register this service to the local IAS service
	// In creating the object, we pass the name and the length of the
	// name. For attributes, we will only register one attribute
	// namely the local port number using the correct attribute name.
	reqTTPConnection = connectionless;
	
	ias_object.name = (unsigned char*)name;
	ias_object.len = StrLen((const char*)name);
	ias_object.nAttribs = 1;
	ias_object.attribs = &attrib;
	
	// Populate the attribute structure
	if (reqTTPConnection)
	{
		attrib.name = (unsigned char*)attr_lmp_name;
		IrSetConTypeLMP(&con);
	}
	else
	{
		attrib.name = (unsigned char*)attr_ttp_name;
		IrSetConTypeTTP(&con);
	}
	
	attrib.len = StrLen((const char*)attrib.name);
	attrib.value = (unsigned char*)attr_value;
	attrib.valLen = 5;	
	attr_value[4] = con.lLsap;
	
	
	// Now add the ias object we just created to the database.
	if(IrIAS_Add(refNum, &ias_object) != IR_STATUS_SUCCESS)
	{
		debug_msg("IrOpenComm::IrIasAdd failed.");
		IR_CloseCom();
		return IR_STATUS_FAILED;
	}
	
	debug_msg("IR_Bind successful");
	
	return IR_STATUS_SUCCESS;
}



/***********************************************************************
 *
 * FUNCTION:     IR_Connect
 *
 * DESCRIPTION:  This function will make an actual connection to another
 * device. This will attempt to find the first device it can and connect
 * to it.
 *
 * **NOTE** You must call IR_OpenCom first before calling this function
 *          as you need to open the IR com port first.
 *
 * PARAMETERS:   
 *
 * RETURNED:     IR_STATUS_SUCCESS | IR_STATUS_FAILED;
 *
 * REVISION HISTORY:
 *
 ***********************************************************************/
IrStatus IR_Connect()
{
	IrStatus err;
	IrPacket packet;
	UInt16 iCounter = 0;
	Boolean leave = false;
	
	// If we are connected, then we can exit.
	if (connected == true)
		return IR_STATUS_SUCCESS;
		
	// We must check first to see if we have bound to the IR com.
	if (refNum == 0)
	{
		debug_msg("Must Open the IR port first");
		return IR_STATUS_FAILED;
	}
	
	// We are bounded to the IR port, we now must discover another user.
	while (iCounter++ < DiscTimeout && !devAddrValid)
	{
		err = IrDiscoverReq(refNum, &con);
		
		switch (err)
		{
			case IR_STATUS_MEDIA_BUSY:
			{
				debug_msg("IrOpenComm::IrDiscoverReq is busy");
				leave = true;
				break;
			}
		
			case IR_STATUS_PENDING:
				debug_msg("IrDiscoverReq executed.");
				leave = true;
				break;
				
			default:
				debug_msg("Error occurred in IrOpenComm::IrDiscoverReq");
				IR_CloseCom();
				return IR_STATUS_FAILED;
		}
		
		if (leave)
			break;
	}
	
	iCounter = 0;
	while (iCounter++ < DiscTimeout && devAddrValid == false);
	
	if (!devAddrValid)
	{
		debug_msg("Failed to make an IR connection.");
		return IR_STATUS_FAILED;
	}
		
	
	// Now attempt to connect to an IrLap on their end.
	
	iCounter = 0;
	while (iCounter++ < DiscTimeout && !irLapSuccess)
	{
		err = IrConnectIrLap(refNum, device_addr);
		
		switch (err)
		{
			case IR_STATUS_MEDIA_BUSY:
				debug_msg("IrOpenComm::IrConnectIrLap media is busy.");
				break;
			
			case IR_STATUS_PENDING:
				debug_msg("IrOpenComm::IrConnectIrLap executed.");
				leave = true;
				break;
				
			default:
				debug_msg("Error occurred in IrOpenComm::IrConnectIrLap");
				IR_CloseCom();
				return IR_STATUS_FAILED;
		}
		
		if (leave)
			break;
	}
	
	for (iCounter = 0; iCounter < DiscTimeout && irLapSuccess == false; iCounter++);
	
	if (!irLapSuccess)
	{
		debug_msg("Unable to make a successful IrLap");
		debug_msg("connection. Timed out");
		return IR_STATUS_FAILED;
	}
	
	
	// We now have a connection. We now must query the other's IAS to find the
	// LAP port to make the final connection.
	if (!gotRLsap)
	{
		iasQuery.queryBuf = (reqTTPConnection) ? irTTPQuery : irLMPQuery;
		iasQuery.queryLen = (reqTTPConnection) ? irTTPQueryLen : irLMPQueryLen;
		iasQuery.callBack = ias_cb_func;
		iasQuery.result = qRsltBuff;
		iasQuery.resultBufSize = qRsltBuffSize;
		
		IrIAS_StartResult(&iasQuery);
		err = IrIAS_Query(refNum, &iasQuery);
		if (err != IR_STATUS_SUCCESS)
		{
			debug_msg("Unable to query IAS for RLsap");
			return IR_STATUS_FAILED;
		}
	}
	
	for (iCounter = 0; iCounter < DiscTimeout && gotRLsap == false; iCounter++);
	
	if (!gotRLsap)
	{
		debug_msg("Unable to obtain RLsap. Query Failed to return");
		return IR_STATUS_FAILED;
	}
	
	
	// Before calling this function, you should set the type of connection
	// you want (either TTP or LMP) but we already did this in the
	// IR_OpenCom function when we were setting up the IAS DB entry.
	packet.buff = NULL;
	packet.len = 0;
	
	err = IrConnectReq(refNum, &con, &packet, 0);
	
	if (err != IR_STATUS_PENDING)
	{
		debug_msg("Unable to establish a connection");
		return IR_STATUS_FAILED;
	}
	
	// At this point, we are complete and ready to send.
	debug_msg("Ready to send");
	
	return IR_STATUS_SUCCESS;
}


/***********************************************************************
 *
 * FUNCTION:     IR_SendPacket
 *
 * DESCRIPTION:  Use this function to send your packets. You must have
 * already open the IR port and made a connection to another device
 * before continuing.
 *
 * PARAMETERS:   IrPacket* packet -- the packet you want to send.
 *				 Byte credit     -- The amount of credit you want
 *								     to send. See docs for more info.
 *
 * RETURNED:     IR_STATUS_SUCCESS | (err result of IrDataReq)
 *
 * REVISION HISTORY:
 *
 ***********************************************************************/
IrStatus IR_SendPacket(IrPacket* packet) {
	
	IrStatus err;
	
	if (!connected)
	{
		debug_msg("Must connect first");
		return IR_STATUS_FAILED;
	}
	
	err = IrDataReq(refNum, &con, packet);
	
	return (err == IR_STATUS_PENDING) ? IR_STATUS_SUCCESS : err;
}


IrStatus IR_ReceiveConnection()
{
	IrStatus err;
	IrPacket packet;
	
	if (!refNum)
	{
		debug_msg("You must bind first.");
		return IR_STATUS_FAILED;
	}
	
	packet.len = 0;
	packet.buff = NULL;
	
	err = IrConnectRsp(refNum, &con, &packet, 0);
	
	if (err == IR_STATUS_PENDING)
		return IR_STATUS_SUCCESS;
		
	if (err == IR_STATUS_FAILED)
		debug_msg("IR_ReceiveConn::Failed.");
	else
		debug_msg("IR_ReceiveConn::No IrLap");
		
	return IR_STATUS_FAILED;
}

/***********************************************************************
 *
 * FUNCTION:     IR_Close
 *
 * DESCRIPTION:  This function will tear down the appropriate connections
 * for the IR communication
 *
 * **NOTE** You must call IR_Close when your application terminates or
 * shuts down. If you do not, the IR port will remain open. Unexpected
 * results may occur.
 *
 * PARAMETERS:   
 *
 * RETURNED:     
 *
 * REVISION HISTORY:
 *
 ***********************************************************************/
void IR_CloseCom()
{
	if (refNum != 0)
		IrDisconnectIrLap(refNum);
		
	if (refNum != 0)
		IrUnbind(refNum, &con);
	
	if (refNum != 0)
		IrClose(refNum);

	// Now reset internal states.		
	refNum = 0;
	connected = false;
	irLapSuccess = false;
	gotRLsap = false;
	devAddrValid = false;
	foundObject = false;
}