/**
 * Author: 	Kevin Chan
 *
 * Acknowledgements:
 *		Portions of this code was taken from:
 *          Till Harbaum
 *
 *			t.harbaum@tu-bs.de
 *			http://www.ibr.cs.tu-bs.de/
 *
 * Description:
 * 		The code present here is the main loop for the UI of our
 *		chat program.
 */
#include <MemoryMgr.h>
#include <SystemMgr.h>

#include "ir-chat_ui_res.h"
#include "AppStdIO.h"
#include "serial_ir.h"

static ControlType* btnSend = NULL;
static FieldType* fldP = NULL;
static Boolean connected = false;



/***********************************************************************
 *
 * FUNCTION:     GetObjectPtr
 *
 * DESCRIPTION:  Given an object ID and the containing form, it will
 * return a pointer to the object.
 *
 * PARAMETERS:   -->frmPtr  // A form ptr containing the ObjectID we 
 *							// are looking for.
 *
 *				 -->objID	// The Object ID of the object we want.
 *
 * RETURNED:     MemPtr: Pointer to the object that was requested.
 *						 ** NULL ** if it is an error.
 *
 * REVISION HISTORY:
 *
 ***********************************************************************/
static MemPtr GetObjectPtr(FormType * const frmPtr, UInt16 objID)
{

	UInt16 index = FrmGetObjectIndex(frmPtr, objID);
	return FrmGetObjectPtr(frmPtr, index);
}
 

/***********************************************************************
 *
 * FUNCTION:     StartApp
 *
 * DESCRIPTION:  This is called when the application is starting up.
 * At this point, we need to initialize our standard io and to startup
 * the main form.
 *
 * PARAMETERS:   nothing
 *
 * RETURNED:     true if there is an error starting the application.
 *
 * REVISION HISTORY:
 *
 ***********************************************************************/
static void StartApp()
{
	
	StdInit(MainForm, 
  		  mainStdOutField,
  		  mainStdOutScrollBar,
  		  NULL);
  	
  	FrmGotoForm(MainForm);
  	
  	
  	IrOpenPort(9600);
  	
  	// Default settings are fine.
  	// Change here if you need to.
  	// IrSetSerialSettings();
  	
  	debug_msg("Application Started");
}

/***********************************************************************
 *
 * FUNCTION:     StopApplication
 *
 * DESCRIPTION:  This is called when the application is shutting down.
 * Do any last minute cleanups here.
 *
 * PARAMETERS:   nothing
 *
 * RETURNED:     true if there is an error starting the application.
 *
 * REVISION HISTORY:
 *
 ***********************************************************************/
static void StopApp()
{
	StdFree();
}


/***********************************************************************
 *
 * FUNCTION:     MainFormEventHandler
 *
 * DESCRIPTION:  This is the main event handler taht is associated with.
 * out main form;
 *
 * PARAMETERS:   --> event  // The event that we may handle.
 *
 * RETURNED:     true if there is an error starting the application.
 *
 * REVISION HISTORY:
 *
 ***********************************************************************/
static Boolean MainFormEventHandler(EventPtr event)
{
	Boolean handled = false; 
	if(StdHandleEvent(event)) return true;
  
	switch (event->eType) {
	  	case frmOpenEvent:
		    FrmDrawForm(FrmGetActiveForm());			
		    handled = true;
		    break;
	}
	
	return(handled);
}


/***********************************************************************
 *
 * FUNCTION:     AppEventHandler
 *
 * DESCRIPTION:  This is the main event handler for our chat program.
 *
 * PARAMETERS:	--> event  // The event that we may handle.
 *
 * RETURNED:     true if there is an error starting the application.
 *
 * REVISION HISTORY:
 *
 ***********************************************************************/
static Boolean AppEventHandler(EventType* event)
{
	FormPtr ptrForm;
	UInt16    formId;
	Boolean handled = false;
	
	switch (event->eType)
	{
		case frmLoadEvent:
		{
			formId = event->data.frmLoad.formID;
			ptrForm = FrmInitForm(formId);
			FrmSetActiveForm(ptrForm);
		
			switch (formId) {
			case MainForm:
				// Register the form event handler for the Main Form.
				FrmSetEventHandler(ptrForm, MainFormEventHandler);
			  
				// Grab the btn Object Ptrs.
				btnSend = GetObjectPtr(ptrForm, mainBtnSendButton);
			  	fldP = GetObjectPtr(ptrForm, mainTxtPreviewField);
			  
			 
			  break;
			}
			
			handled = true;
			break;
		}
	
		case appStopEvent:
		{
			IrClosePort();
			handled = true;
			
			// Cleanup the Main Form's data.
			// We need to free the forms information,
			// this is better than allowing it to be
			// garbage collected
			ptrForm = FrmGetFormPtr(MainForm);
			if (ptrForm)
			{
				FrmEraseForm(ptrForm);
				FrmDeleteForm(ptrForm);
			}
				
			break;
		}
		
		case ctlEnterEvent:
		{
			switch(event->data.ctlEnter.controlID)
			{					
				case mainBtnSendButton:
				{
					MemHandle stringH;
					char* textP;
					UInt32 bytesSent;
					UInt16 textLength;
					
					handled = true;
					
					stringH = FldGetTextHandle(fldP);
					textLength = FldGetTextLength(fldP);
					
					if (textLength == 0) 
						break;
					
					if (!stringH)
						break;
					
					// If we grab the memory handle, we will get
					// a hold of the internal text.
					textP = (char*)MemHandleLock(stringH);
					
					
					
					if (!IrSendData(textP, textLength, &bytesSent));
						DISPLAY_SND_MSG(textP);
						
					MemHandleUnlock(stringH);
					
					// Now we want to empty the text box.
					FldDelete(fldP, 0, FldGetTextLength(fldP));
					
					break;
				}
					
				default:
					break;
			}
		}
		
		case nilEvent:
	    {
	    	// The nilEvent is a side effect of a number of things.
	    	// In the base case, it is returned by EvtGetEvent routine
	    	// in our event loop. See EvtGetEvent in the documentation
	    	// for more details.
	    	//
	    	// When a nil event arrives, nothing we are assuming that
	    	// nothing is actively running on the palm. So this is a 
	    	// good time to open the serial port and attempt to read
	    	// data.
	    	static UInt32 size = MAX_RCV_QUEUE_SIZE;
	    	static unsigned char* data[MAX_RCV_QUEUE_SIZE];		// Use a static buffer to avoid dynamic allocations.
	    	UInt32 bytesRcvd = 0, bytesLeft = 0;
	    	UInt16 lineErr = 0;
	    	Err err = 0;
	 
 		   		    		    	
    		// We want to yank out all of the data.
    		do {
    		
	    		err = IrReceiveData(data, size - 1,  &bytesRcvd, SysTicksPerSecond() / 5, &lineErr);
	    			
	    		if (err == 0)
	    		{
	    			if (bytesRcvd > 0)
	    			{	    				
	    				// Set the null pointer. Valid because we told
	    				// IrReceiveData that buffer was size - 1;
		    			((char*)data)[bytesRcvd] = '\0';
		    			
		    			DISPLAY_RCV_MSG((char*)data);
		    		}
	    		}
	    		
    		} while ((IrReceiveCheck(&bytesLeft, NULL) == 0) && bytesLeft > 0);

			handled = true;
			break;
		}
		
	}
	
	return handled;
}


/***********************************************************************
 *
 * FUNCTION:     EventLoop
 *
 * DESCRIPTION:  This is where we retrieve and process the events
 *               that are happening in the system.
 *
 * PARAMETERS:
 *
 * RETURNED:     true if there is an error starting the application.
 *
 * REVISION HISTORY:
 *
 ***********************************************************************/
static void EventLoop()
{
	EventType event;
    UInt16      error;

    do
    {
    	// Retrieve an event from the event queue.
        EvtGetEvent(&event, SysTicksPerSecond() * 4);
        
        /* Place any "must handle events before system" logic here */
        
        /* -------------------------------- */
        
        
        // Now allow the System a chance to handle any events that it
        // needs to.
        if (SysHandleEvent(&event))
            continue;
            
        // After the system, we will handle the events using our
        // own event handler for the entire application.
        if (AppEventHandler(&event))
            continue;
            
        // This event handler is used to catch a btn click on the
        // btnConnDiscon button. (which is initialized in frmLoadEvent            
        if (btnSend && CtlHandleEvent(btnSend, &event))
        	continue;
        	
        // Default processing of events, if we
        // do not handle them.
        FrmDispatchEvent(&event);
    }
    while (event.eType != appStopEvent);
}


/***********************************************************************
 *
 * FUNCTION:     Pilot Main
 *
 * DESCRIPTION:  This is the main entry point into the application.
 *
 * PARAMETERS:		// See specification for parameter details.
 *
 * RETURNED:     true if there is an error starting the application.
 *
 * REVISION HISTORY:
 *
 ***********************************************************************/
UInt32 PilotMain(UInt16 cmd, MemPtr cmdBPB, UInt16 launchFlags)
{
	if (cmd == sysAppLaunchCmdNormalLaunch)
    {
        StartApp();
        EventLoop();
        StopApp();
        return(0);
    }
    return(0);
}